/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2016 Jan Kelling
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

///\file
///\brief This files provides an c++14 std::make_unique implementation for c++11 compilers.

#pragma once

#include <memory>
#include <functional>
#include <type_traits>

namespace nytl
{

#if __cplusplus >= 201402L
 using std::make_unique;
#else

namespace detail
{

template<typename T> struct uniqueType { using singleObject = std::unique_ptr<T>; };
template<typename T> struct uniqueType<T[]> { using unknownBound = std::unique_ptr<T[]>; };
template<typename T, std::size_t N> struct uniqueType<T[N]> { using knownBound = void; };

}

template<typename T, typename... Args >
typename detail::uniqueType<T>::singleObject make_unique(Args&&... args)
{
    return std::unique_ptr<T>(new T(std::forward<Args>(args)...));
}

template<typename T>
typename detail::uniqueType<T>::unknownBound make_unique(std::size_t size)
{
    return std::unique_ptr<T>(new typename std::remove_extent<T>::type[size]());
}

template<typename T, typename... Args>
typename detail::uniqueType<T>::knownBound make_unique(std::size_t size) = delete;

#endif // __cplusplus
}
